import abc
import asyncio
import ipaddress
import re
import sys
from json import dumps as json_dumps

import idna
import aiodns

from config import DEFAULT_TIMEOUT
from providers import Provider, BASE_PROVIDERS, DNSBL_CATEGORY_ERROR

if sys.platform == 'win32' and sys.version_info >= (3, 8):
    # fixes https://github.com/dmippolitov/pydnsbl/issues/12
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class DNSBLResult:
    """
    DNSBL Result class to keep all info about ip request results.

    Attributes:
        * request - checked ip/domain
        * providers - dnsbl that was asked for response while checking
        * failed_provider - dnsbl that was unable to provide result due
            to connection issues (connection timeout etc...)
        * detected_by - list of providers that have the ip listed
        * provider_categories - dnsbl that have ip listed and categories detected by
            this dnsbls. dict: {'dnsbl_list_name': list(categories_from_this_dnsbl)}
        * categories - set of dnsbl categories from all providers (subset of DNSBL_CATEGORIES)
    """
    def __init__(self, request: str, results: any):
        self.request = request
        self._results = results
        self.detected = False
        self.providers: list[Provider] = []
        self.failed_providers: list[Provider] = []
        self.detected_by: list[Provider] = []
        self.provider_categories: dict[str: list[str]] = {}
        self.categories = set()
        self.process_results()

    def process_results(self):
        for result in self._results:
            if not hasattr(result, 'provider'):
                self.failed_providers.append(result)
                continue

            provider = result.provider
            self.providers.append(provider)
            if result.error:
                self.failed_providers.append(provider)
                continue

            if not result.response:
                continue

            # set detected to True if ip is detected with at least one dnsbl
            provider_categories = provider.process_response(result.response)
            # If the response is an error, do not consider it as detected
            # (refer to https://www.spamhaus.org/faqs/domain-blocklist/#291:~:text=The%20following%20special%20codes%20indicate%20an%20error)
            if provider_categories != {DNSBL_CATEGORY_ERROR}:
                self.detected = True
                self.categories = self.categories.union(provider_categories)
                self.detected_by.append(provider)
                self.provider_categories[provider.host] = list(provider_categories)

    def __repr__(self):
        detected = ' [DETECTED]' if self.detected else ''
        return f"<DNSBLResult: {self.request}{detected} ({len(self.detected_by)}/{len(self.providers)})>"

    def to_dict(self) -> dict:
        return {
            'request': self.request,
            'detected': self.detected,
            'detected_by': [p.host for p in self.detected_by],
            'categories': list(self.categories),
            'count': {
                'detected': len(self.detected_by),
                'checked': len(self.providers),
                'failed': len(self.failed_providers),
            },
            'detected_provider_categories': self.provider_categories,
            'checked_providers': [p.host for p in self.providers],
            'failed_providers': [p.host for p in self.failed_providers],
        }

    def to_json(self, indent: int = 2) -> str:
        return json_dumps(self.to_dict(), indent=indent)


class DNSBLResponse:
    def __init__(self, request: str, provider: Provider, response: any, error: any):
        self.request = request
        self.provider = provider
        self.response = response
        self.error = error


class BaseAsyncDNSBLChecker(abc.ABC):
    def __init__(self, providers: list[Provider] = BASE_PROVIDERS, timeout = DEFAULT_TIMEOUT):
        self.providers: list[Provider] = []
        self._timeout = timeout
        for p in providers:
            if not hasattr(p, 'host'):
                raise ValueError(f'providers should contain only Provider instances: {p} {type(p)}')

            self.providers.append(p)

        self._resolver = None

    async def __aenter__(self):
        self._resolver = aiodns.DNSResolver(timeout=self._timeout)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._resolver:
            await self._resolver.close()

    async def query_provider(self, request: str, provider: Provider) -> DNSBLResponse:
        response = None
        error = None
        dnsbl_query = f"{self.prepare_query(request)}.{provider.host}"

        try:
            response = await self._resolver.query(dnsbl_query, 'A')

        except aiodns.error.DNSError as exc:
            if exc.args[0] != 4: # 4: domain name not found:
                error = exc

        return DNSBLResponse(request=request, provider=provider, response=response, error=error)

    @abc.abstractmethod
    def prepare_query(self, request: str):
        return NotImplemented

    async def check(self, request: str) -> DNSBLResult:
        tasks = []
        for provider in self.providers:
            tasks.append(self.query_provider(request, provider))

        results = await asyncio.gather(*tasks, return_exceptions=True)
        return DNSBLResult(request=request, results=results)

    async def bulk_check(self, requests: list[str]) -> list[DNSBLResult]:
        results = []
        for request in requests:
            results.append(await self.check(request))

        return results


class AsyncCheckIP(BaseAsyncDNSBLChecker):
    def prepare_query(self, request):
        address = ipaddress.ip_address(request)
        if address.version == 4:
            return '.'.join(reversed(request.split('.')))

        if address.version == 6:
            # according to RFC: https://tools.ietf.org/html/rfc5782#section-2.4
            request_stripped = address.exploded.replace(':', '')
            return '.'.join(reversed(request_stripped))

        raise ValueError('unknown ip version')


class AsyncCheckDomain(BaseAsyncDNSBLChecker):
    # https://regex101.com/r/vdrgm7/1
    DOMAIN_REGEX = re.compile(r"^(((?!-))(xn--|_{1,1})?[a-z0-9-]{0,61}[a-z0-9]{1,1}\.)*(xn--[a-z0-9][a-z0-9\-]{0,60}|[a-z0-9-]{1,30}\.[a-z]{2,})$")

    def prepare_query(self, request):
        request = request.lower() # Adding support for capitalized letters in domain name.
        domain_idna = idna.encode(request).decode()
        if not self.DOMAIN_REGEX.match(domain_idna):
            raise ValueError(f'should be valid domain, got {domain_idna}')

        return domain_idna


class BaseDNSBLChecker:
    def __init__(self,
             async_checker: BaseAsyncDNSBLChecker,
             providers: list[Provider] = BASE_PROVIDERS, timeout = DEFAULT_TIMEOUT,
     ):
        self._async_checker = async_checker(providers=providers, timeout=timeout)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del exc_tb, exc_val, exc_type

    @staticmethod
    def _raise_result_exception(result: DNSBLResult) -> None:
        if isinstance(result, Exception):
            raise result

    async def _check_async(self, request: str) -> DNSBLResult:
        async with self._async_checker as checker:
            return await checker.check(request)

    def check(self, request: str) -> DNSBLResult:
        result = asyncio.run(self._check_async(request))
        self._raise_result_exception(result)
        return result

    async def _bulk_check_async(self, request: list[str]) -> list[DNSBLResult]:
        async with self._async_checker as checker:
            return await checker.bulk_check(request)

    def bulk_check(self, request: list[str]) -> list[DNSBLResult]:
        results = asyncio.run(self._bulk_check_async(request))
        for r in results:
            self._raise_result_exception(r)

        return results


class CheckIP(BaseDNSBLChecker):
    def __init__(self, providers: list[Provider] = BASE_PROVIDERS, timeout = DEFAULT_TIMEOUT):
        BaseDNSBLChecker.__init__(self, async_checker=AsyncCheckIP, providers=providers, timeout=timeout)


class CheckDomain(BaseDNSBLChecker):
    def __init__(self, providers: list[Provider] = BASE_PROVIDERS, timeout = DEFAULT_TIMEOUT):
        BaseDNSBLChecker.__init__(self, async_checker=AsyncCheckDomain, providers=providers, timeout=timeout)



RAW_BASE_PROVIDERS = [
    'all.s5h.net',
    'aspews.ext.sorbs.net',
    'b.barracudacentral.org',
    'bl.nordspam.com',
    'blackholes.five-ten-sg.com',
    'blacklist.woody.ch',
    # The provider zen.spamhaus.org is already being used. abuseat.org redirects to spamhaus.org
    # Additionally, abuseat.org has the same behaviour as zen.spamhaus.org
    # and we manage the new DNSBL_CATEGORY_ERROR in the zen.spamhaus.org Provider class
    # 'cbl.abuseat.org',
    'combined.abuse.ch',
    'combined.rbl.msrbl.net',
    'db.wpbl.info',
    'dnsbl.cyberlogic.net',
    'dnsbl.sorbs.net',
    'drone.abuse.ch',
    'images.rbl.msrbl.net',
    'ips.backscatterer.org',
    'ix.dnsbl.manitu.net',
    'korea.services.net',
    'matrix.spfbl.net',
    'phishing.rbl.msrbl.net',
    'proxy.bl.gweep.ca',
    'proxy.block.transip.nl',
    'psbl.surriel.com',
    'rbl.interserver.net',
    'relays.bl.gweep.ca',
    'relays.bl.kundenserver.de',
    'relays.nether.net',
    'residential.block.transip.nl',
    'singular.ttk.pte.hu',
    'spam.dnsbl.sorbs.net',
    'spam.rbl.msrbl.net',
    'spambot.bls.digibase.ca',
    'spamlist.or.kr',
    'spamrbl.imp.ch',
    'spamsources.fabel.dk',
    'ubl.lashback.com',
    'virbl.bit.nl',
    'virus.rbl.msrbl.net',
    'virus.rbl.jp',
    'wormrbl.imp.ch',
    'z.mailspike.net',
]
RAW_DOMAIN_PROVIDERS = [
    'uribl.spameatingmonkey.net',
    'multi.surbl.org',
    'rhsbl.sorbs.net '
]
DNSBL_CATEGORY_UNKNOWN = 'unknown'
DNSBL_CATEGORY_SPAM = 'spam'
DNSBL_CATEGORY_EXPLOITS = 'exploits'
DNSBL_CATEGORY_PHISH = 'phish'
DNSBL_CATEGORY_MALWARE = 'malware'
DNSBL_CATEGORY_CNC = 'cnc'
DNSBL_CATEGORY_ABUSED = 'abused'
DNSBL_CATEGORY_LEGIT = 'legit'
DNSBL_CATEGORY_ERROR = 'error'
DEFAULT_TIMEOUT = 5



"""
Place to define providers.
Most part of _BASE_PROVIDERS was taken from https://github.com/vincecarney/dnsbl
"""
### DNSBL CATEGORIES ###
# providers answers could be interpreted in one of the following categories

# pylint: disable=W0401,W0614
from config import *


class Provider:
    def __init__(self, host: str):
        self.host = host

    def process_response(self, response):
        """
        Usually DNSBL lists returns ip-codes like this: 127.0.0.2
        Some of the lists provides some specification about
        this codes and their meaning. This function will be helpful to build mapping
        between response codes and DNSBL_CATEGORIES.  It is used in construction
        of DNSBLResult. You should redefine this function
        in your custom providers according to their specification.

        Parmeters:
            result - list of c-ares dns responses

        Returns:
            set of categories (DNSBL_CATEGORIES subset)

        """
        result = set()
        if response:
            result.add(DNSBL_CATEGORY_UNKNOWN)

        return result

    def __repr__(self):
        return f"<Provider: {self.host}>"


class ZenSpamhaus(Provider):
    """ Combined spamhaus list:
        https://www.spamhaus.org/faq/section/DNSBL%20Usage#200
    """

    def __init__(self):
        Provider.__init__(self, host='zen.spamhaus.org')

    def process_response(self, response):
        categories = set()
        for result in response:
            if result.host in ['127.0.0.2', '127.0.0.3', '127.0.0.9']:
                categories.add(DNSBL_CATEGORY_SPAM)

            elif result.host in ['127.0.0.4', '127.0.0.5', '127.0.0.6', '127.0.0.7']:
                categories.add(DNSBL_CATEGORY_EXPLOITS)

            elif result.host in ['127.255.255.252', '127.255.255.254', '127.255.255.255']:
                categories.add(DNSBL_CATEGORY_ERROR)

            else:
                categories.add(DNSBL_CATEGORY_UNKNOWN)

        return categories


class DblSpamhaus(Provider):
    """ Spamhaus domain blacklist
        https://www.spamhaus.org/faq/section/Spamhaus%20DBL#291
    """
    CATEGORY_MAPPING = {
        '127.0.1.2': {DNSBL_CATEGORY_SPAM},
        '127.0.1.4': {DNSBL_CATEGORY_PHISH},
        '127.0.1.5': {DNSBL_CATEGORY_MALWARE},
        '127.0.1.6': {DNSBL_CATEGORY_CNC},
        '127.0.1.102': {DNSBL_CATEGORY_ABUSED, DNSBL_CATEGORY_LEGIT, DNSBL_CATEGORY_SPAM},
        '127.0.1.103': {DNSBL_CATEGORY_ABUSED, DNSBL_CATEGORY_SPAM},
        '127.0.1.104': {DNSBL_CATEGORY_ABUSED, DNSBL_CATEGORY_LEGIT, DNSBL_CATEGORY_PHISH},
        '127.0.1.105': {DNSBL_CATEGORY_ABUSED, DNSBL_CATEGORY_LEGIT, DNSBL_CATEGORY_MALWARE},
        '127.0.1.106': {DNSBL_CATEGORY_ABUSED,  DNSBL_CATEGORY_LEGIT, DNSBL_CATEGORY_CNC},
        '127.255.255.252': {DNSBL_CATEGORY_ERROR},
        '127.255.255.254': {DNSBL_CATEGORY_ERROR},
        '127.255.255.255': {DNSBL_CATEGORY_ERROR},
    }

    def __init__(self):
        Provider.__init__(self, host='dbl.spamhaus.org')

    def process_response(self, response):
        categories = set()
        for result in response:
            result_categories = self.CATEGORY_MAPPING.get(result.host, {DNSBL_CATEGORY_UNKNOWN})
            categories.update(result_categories)

        return categories


BASE_PROVIDERS = [Provider(host) for host in RAW_BASE_PROVIDERS] + [ZenSpamhaus()]
BASE_DOMAIN_PROVIDERS = [Provider(host) for host in RAW_DOMAIN_PROVIDERS] + [DblSpamhaus()]



import pytest


def test_1():
    from providers import Provider
    from checker import CheckIP, CheckDomain, AsyncCheckDomain
    pass

# todo: mock responses of dns-resolver to get stable output
# todo: CheckIP
# todo: CheckDomain
# todo: invalid IPs and Domains
# todo: Domain variants (capitalized / 3th+ levels, idna)
# todo: IPv4 + IPv6


# def test_ipv6_converting():
#     # https://datatracker.ietf.org/doc/html/rfc5782#section-2.4
#     checker = AsyncCheckDomain()
#     assert checker.prepare_query('2001:db8:1:2:3:4:567:89ab') == "b.a.9.8.7.6.5.0.4.0.0.0.3.0.0.0.2.0.0.0.1.0.0.0.8.b.d.0.1.0.0.2"
#     assert checker.prepare_query('2600:2600::f03c:91ff:fe50:d2') == "2.d.0.0.0.5.e.f.f.f.1.9.c.3.0.f.0.0.0.0.0.0.0.0.0.0.6.2.0.0.6.2"



from os import path as os_path
from sys import path as sys_path

sys_path.append(os_path.dirname(os_path.abspath(__file__)))

# pylint: disable=C0413
from checker import CheckIP, CheckDomain

__all__ = (
    "CheckIP",
    "CheckDomain",
)



from os import path as os_path
from sys import path as sys_path
from argparse import ArgumentParser
from json import dumps as json_dumps

sys_path.append(os_path.dirname(os_path.abspath(__file__)))

# pylint: disable=C0413
from checker import CheckDomain, CheckIP


def main():
    parser = ArgumentParser(
        prog='DNS-BL Lookup-Client'
    )
    g = parser.add_mutually_exclusive_group(required=True)
    g.add_argument('-i', '--ip', type=str, default=None, help='IP to check')
    g.add_argument('-d', '--domain', type=str, default=None, help='Domain to check')
    parser.add_argument('-j', '--json', type=bool, default=False, help='Only output JSON')
    parser.add_argument(
        '--details', action='store_true', default=False,
        help='If the result details should be added to the output',
    )
    args = parser.parse_args()

    if args.ip is not None:
        if not args.json:
            print(f'Checking IP {args.ip} ..')

        with CheckIP() as checker:
            result = checker.check(args.ip)

    else:
        if not args.json:
            print(f'Checking Domain {args.domain} ..')

        with CheckDomain() as checker:
            result = checker.check(args.domain)

    response = {
        'detected': result.detected,
        'detected_by': [p.host for p in result.detected_by],
        'categories': list(result.categories),
        'count': {
            'detected': len(result.detected_by),
            'checked': len(result.providers),
            'failed': len(result.failed_providers),
        }
    }
    if args.details:
        response['detected_provider_categories'] = result.provider_categories
        response['checked_providers'] = [p.host for p in result.providers]
        response['failed_providers'] = [p.host for p in result.failed_providers]

    print(json_dumps(response, indent=2))


if __name__ == '__main__':
    main()
